#import "Character.h"
#import <stdlib.h>

//
// Experience table
//

int levels[] = {
	0,		1000,	2250,	3750,	5500,	7500,	10000,	13000,	16500,	20500,	//Heroic tier
	26000,	32000,	39000,	47000,	57000,	69000,	83000,	99000,	119000,	143000,	//Paragon tier
	175000,	210000,	255000,	310000,	375000,	450000,	550000,	675000,	825000,	1000000	//Epic tier
};

@implementation Character

/*******************************
 * Static & pseudostatic stuff *
 *******************************/

+ (void)initialize { //specify dependent keys
	[super initialize];

	// abilities 
	[self setKeys:[Character statNames] triggerChangeNotificationsForDependentKey:@"pointCost"];

	//race & class
	[self setKeys:[NSArray arrayWithObjects:@"race", nil] triggerChangeNotificationsForDependentKey:@"raceBonuses"];
	[self setKeys:[NSArray arrayWithObjects:@"cclass", nil] triggerChangeNotificationsForDependentKey:@"classBonuses"];
	
	// xp <-> level
	[self setKeys:[NSArray arrayWithObjects:@"experience", nil] triggerChangeNotificationsForDependentKey:@"level"];
	
	// calculated modifiers
	[self setKeys:[NSArray arrayWithObjects:@"level", nil] triggerChangeNotificationsForDependentKey:@"halfLevel"];
	
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"strength", nil] triggerChangeNotificationsForDependentKey:@"strMod"];
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"constitution", nil] triggerChangeNotificationsForDependentKey:@"conMod"];
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"dexterity", nil] triggerChangeNotificationsForDependentKey:@"dexMod"];
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"intelligence", nil] triggerChangeNotificationsForDependentKey:@"intMod"];
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"wisdom", nil] triggerChangeNotificationsForDependentKey:@"wisMod"];
	[self setKeys:[NSArray arrayWithObjects:@"raceBonuses", @"charisma", nil] triggerChangeNotificationsForDependentKey:@"chaMod"];
	
	[self setKeys:[NSArray arrayWithObjects:@"dexMod", @"intMod", nil] triggerChangeNotificationsForDependentKey:@"ACAbil"];
	[self setKeys:[NSArray arrayWithObjects:@"strMod", @"conMod", nil] triggerChangeNotificationsForDependentKey:@"fortAbil"];
	[self setKeys:[NSArray arrayWithObjects:@"dexMod", @"intMod", nil] triggerChangeNotificationsForDependentKey:@"refAbil"];
	[self setKeys:[NSArray arrayWithObjects:@"wisMod", @"chaMod", nil] triggerChangeNotificationsForDependentKey:@"willAbil"];
	
	[self setKeys:[NSArray arrayWithObjects:@"halfLevel", @"dexMod", nil] triggerChangeNotificationsForDependentKey:@"initiative"];
	
	// passive values
	[self setKeys:[NSArray arrayWithObjects:@"ACAbil", @"halfLevel", @"classBonuses", nil] triggerChangeNotificationsForDependentKey:@"AC"];
	[self setKeys:[NSArray arrayWithObjects:@"fortAbil", @"halfLevel", @"classBonuses", nil] triggerChangeNotificationsForDependentKey:@"fort"];
	[self setKeys:[NSArray arrayWithObjects:@"refAbil", @"halfLevel", @"classBonuses", nil] triggerChangeNotificationsForDependentKey:@"ref"];
	[self setKeys:[NSArray arrayWithObjects:@"willAbil", @"halfLevel", @"classBonuses", nil] triggerChangeNotificationsForDependentKey:@"will"];
	
	[self setKeys:[NSArray arrayWithObjects:@"constitution", @"classBonuses", @"level", nil] triggerChangeNotificationsForDependentKey:@"hitPoints"];
	[self setKeys:[NSArray arrayWithObjects:@"conMod", @"classBonuses", nil] triggerChangeNotificationsForDependentKey:@"healingSurges"];
	
	// raw XML
	[self setKeys:[NSArray arrayWithObjects:
				   @"AC", @"fort", @"ref", @"will",
				   @"classBonuses", @"raceBonuses",
				   @"name", @"experience",
				   nil] triggerChangeNotificationsForDependentKey:@"data"];
}

+ (int)modifierForValue:(int)stat {
	int mod;
	
	stat -= 10;
	
	if (stat < 0) {
		stat--;
	}
	
	mod = stat / 2;	
	
	return mod;
}		

//Stat validator
- (BOOL)validateValue:(id *)ioValue forKey:(NSString*)key error:(NSError **)outError {
	BOOL validated = false;
	
	if ([[Character statNames] containsObject:key]) {
		if ([*ioValue intValue] >= 3 &&
			[*ioValue intValue] <= 18) {
			validated = true;
		} else {
			*outError = [[NSError alloc] initWithDomain: @"Character data"
												   code: [*ioValue intValue]
											   userInfo: [NSDictionary dictionaryWithObject: @"Stat value out of range."
																					 forKey: NSLocalizedDescriptionKey]];
		}
	} else if ([key isEqual:@"level"]) {
		if ([*ioValue intValue] >= 1 &&
			[*ioValue intValue] <= 30) {
			validated = true;
		} else {
			*outError = [[NSError alloc] initWithDomain: @"Character data"
												   code: [*ioValue intValue]
											   userInfo: [NSDictionary dictionaryWithObject: @"Stat value out of range."
																					 forKey: NSLocalizedDescriptionKey]];
		}		
	} else {
		validated = [super validateValue:ioValue forKey:key error:outError];
	}
	
	return validated;
}





/*******************************
 * Construct, export           *
 *******************************/

- (id)initWithData:(NSData*)sheetXML {
	self = [super init];

	NSError* maybe = [[NSError alloc] init];
	
	sheet = [[NSXMLDocument alloc] initWithData:sheetXML
										options:NSXMLDocumentValidate
										  error:&maybe];

	[self setRaces:[CharacterRace loadRaces]];
	[self setClasses:[CharacterClass loadClasses]];
	
	if (maybe) {
		NSLog(@"Character init error: %@", [maybe localizedDescription]);
	}
	
	return self;
}

- (NSData*)data {
	return [sheet XMLDataWithOptions:NSXMLNodePrettyPrint];
}

+ (NSArray*)statNames {
	return [NSArray arrayWithObjects:@"strength", @"constitution", @"dexterity", @"intelligence", @"wisdom", @"charisma", nil];
}





/*******************************
 * Generative mutators         *
 *******************************/
	
- (int)fdsl {	//Four Dice Strip Lowest
	unsigned int dice[3], lowest;
	int i;

	for (i = 0; i <= 3; i++) {
		dice[i] = (random() % 6) + 1;
	}

	lowest = dice[0];
	for (i = 0; i <= 3; i++) {
		if (dice[i] < lowest)
			lowest = dice[i];
	}
	
	return (dice[0] + dice[1] + dice[2] + dice[3] - lowest);
}

- (void)reroll {
	[self setStrength: [self fdsl]];
	[self setConstitution: [self fdsl]];
	[self setDexterity: [self fdsl]];
	[self setIntelligence: [self fdsl]];
	[self setWisdom: [self fdsl]];
	[self setCharisma: [self fdsl]];
}





/*******************************
 * Calculating accessors       *
 *******************************/

- (int)pointCost {
	int i, l, lowest;
	int stats[6];
	int cost = 0;
	
	stats[0] = [self strength];
	stats[1] = [self constitution];
	stats[2] = [self dexterity];
	stats[3] = [self intelligence];
	stats[4] = [self wisdom];
	stats[5] = [self charisma];
	
	//select lowest stat as sub-10
	lowest = stats[0];
	for (i = 0; i < 6; i++) {
		if (stats[i] <= lowest) {
			l = i;
			lowest = stats[i];
		}
	}
	
	//Add up cost
	for (i = 0; i < 6; i++) {
		if (i == l) {
			switch (stats[i]) {
				case 8:
					cost += 0;
					break;
					
				case 9:
					cost += 1;
					break;
					
				case 10:
					cost += 2;
					break;
					
				case 11:
					cost += 3;
					break;
					
				case 12:
					cost += 4;
					break;
					
				case 13:
					cost += 5;
					break;
					
				case 14:
					cost += 7;
					break;
					
				case 15:
					cost += 9;
					break;
					
				case 16:
					cost += 11;
					break;
					
				case 17:
					cost += 14;
					break;
				
				case 18:
					cost += 18;
					break;
			}
		}
		else {
			switch (stats[i]) {
				case 10:
					cost += 0;
					break;
					
				case 11:
					cost += 1;
					break;
					
				case 12:
					cost += 2;
					break;
					
				case 13:
					cost += 3;
					break;
					
				case 14:
					cost += 5;
					break;
					
				case 15:
					cost += 7;
					break;
				
				case 16:
					cost += 9;
					break;
					
				case 17:
					cost += 12;
					break;
					
				case 18:
					cost += 16;
					break;
			}
		}
	}
	
	return cost;
}

//Stat mod accessors
- (int)strMod {
	return [Character modifierForValue: [self strength] + [self raceBonuses].strength];
}
- (int)conMod {
	return [Character modifierForValue: [self constitution] + [self raceBonuses].constitution];
}
- (int)dexMod {
	return [Character modifierForValue: [self dexterity] + [self raceBonuses].dexterity];
}
- (int)intMod {
	return [Character modifierForValue: [self intelligence] + [self raceBonuses].intelligence];
}
- (int)wisMod {
	return [Character modifierForValue: [self wisdom] + [self raceBonuses].wisdom];
}
- (int)chaMod {
	return [Character modifierForValue: [self charisma] + [self raceBonuses].charisma];
}
- (int)halfLevel {
	return [self level] / 2;
}

//Defenses
- (int)ACAbil {
	int dex = [self dexMod];
	int inte = [self intMod];
	
	if (dex > inte)
		return dex;
	
	return inte;
}
- (int)fortAbil {
	int str = [self strMod];
	int con = [self conMod];
	
	if (str > con)
		return str;
	
	return con;
}
- (int)refAbil {
	int dex = [self dexMod];
	int inte = [self intMod];
	
	if (dex > inte)
		return dex;
	
	return inte;
}
- (int)willAbil {
	int wis = [self wisMod];
	int cha = [self chaMod];
	
	if (wis > cha)
		return wis;
	
	return cha;
}

- (int)AC {
	return 10 + [self ACAbil] + [self halfLevel];
}
- (int)fort {
	return 10 + [self fortAbil] + [self halfLevel] + [self classBonuses].fort;
}
- (int)ref {
	return 10 + [self refAbil] + [self halfLevel] + [self classBonuses].ref;
}
- (int)will {
	return 10 + [self willAbil] + [self halfLevel] + [self classBonuses].will;
}

- (int)hitPoints {
	return [self constitution] + [[self classBonuses] hp] + ([[self classBonuses] hpGain] * ([self level] - 1));
}
- (int)healingSurges {
	return [[self classBonuses] surges] + [self conMod];
}
- (int)initiative {
	return [self halfLevel] + [self dexMod];
}



/*******************************
 * Automatic properties        *
 *******************************/

@synthesize races;
@synthesize classes;

@dynamic raceBonuses;
@dynamic classBonuses;

- (CharacterRace*)raceBonuses {
	return [[self races]
			objectAtIndex:[[self races]
						   indexOfObject:[self race]]];
}

- (CharacterClass*)classBonuses {
	CharacterClass* selectedClass = [[self classes]
									 objectAtIndex:[[self classes]
													indexOfObject:[self cclass]]];
	
	return selectedClass;
}








/*******************************
 * XML doc properties          *
 *******************************/

//Basics
- (NSString*)name {
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/name" error:nil]
					  objectAtIndex:0];
	
	return [loc stringValue];
}
- (void)setName:(NSString*)name {
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/name" error:nil]
					  objectAtIndex:0];

	[loc setStringValue:name];
}

- (NSString*)race {
	NSError* maybe = [[NSError alloc] init];
	
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/race/@name" error:&maybe]
					  objectAtIndex:0];
	
	if (maybe) {
		NSLog(@"-race error: %@", [maybe localizedDescription]);
	}
	
	return [loc stringValue];
}
- (void)setRace:(NSString*)race {
	NSError* maybe = [[NSError alloc] init];

	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/race/@name" error:&maybe]
					  objectAtIndex:0];
	
	if (maybe) {
		NSLog(@"-setRace: error: %@", [maybe localizedDescription]);
	}
	
	[loc setStringValue:race];
}

- (NSString*)cclass {
	NSError* maybe = [[NSError alloc] init];

	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/class/@name" error:&maybe]
					  objectAtIndex:0];
	
	if (maybe) {
		NSLog(@"-cclass error: %@", [maybe localizedDescription]);
	}

	
	return [loc stringValue];
}
- (void)setCclass:(NSString*)cclass {
	NSError* maybe = [[NSError alloc] init];

	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/class/@name" error:&maybe]
					  objectAtIndex:0];
	
	if (maybe) {
		NSLog(@"-setCclass: error: %@", [maybe localizedDescription]);
	}	
	
	[loc setStringValue:cclass];
}


//XP & level
- (int)experience {
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/experience" error:nil]
					  objectAtIndex:0];
	
	return [[loc stringValue] intValue];
}
- (void)setExperience:(int)experience {
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:@"/character/experience" error:nil]
					  objectAtIndex:0];
	
	[loc setStringValue:[NSString stringWithFormat:@"%d",experience]];
}

- (int)level {
	int level;
	int xp = [self experience];
	
	for (level = 0; xp >= levels[level] && level < 30; level++);
	
	return level;
}
- (void)setLevel:(int)level {
	[self setExperience:levels[level - 1]];
}


//Attributes
- (int)attribute:(NSString*)attrName {
	NSString* path = [@"/character/abilityScores/" stringByAppendingString:attrName];
	
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:path error:nil]
					  objectAtIndex:0];
	
	return [[loc stringValue] intValue];
}
- (void)setAttribute:(NSString*)attrName to:(int)attrVal {
	NSString* path = [@"/character/abilityScores/" stringByAppendingString:attrName];
	
	NSXMLNode* loc = [[[sheet rootElement]
					   nodesForXPath:path error:nil]
					  objectAtIndex:0];

	[loc setStringValue:[NSString stringWithFormat:@"%d", attrVal]];
}

- (int)strength {
	return [self attribute:@"strength"];
}
- (void)setStrength:(int)strength {
	[self setAttribute:@"strength" to:strength];
}

- (int)constitution {
	return [self attribute:@"constitution"];
}
- (void)setConstitution:(int)constitution {
	[self setAttribute:@"constitution" to:constitution];
}

- (int)dexterity {
	return [self attribute:@"dexterity"];
}
- (void)setDexterity:(int)dexterity {
	[self setAttribute:@"dexterity" to:dexterity];
}

- (int)intelligence {
	return [self attribute:@"intelligence"];
}
- (void)setIntelligence:(int)intelligence {
	[self setAttribute:@"intelligence" to:intelligence];
}

- (int)wisdom {
	return [self attribute:@"wisdom"];
}
- (void)setWisdom:(int)wisdom {
	[self setAttribute:@"wisdom" to:wisdom];
}

- (int)charisma {
	return [self attribute:@"charisma"];
}
- (void)setCharisma:(int)charisma {
	[self setAttribute:@"charisma" to:charisma];
}


@end
